"""The cimsparql.model module contains the base class CimModel for both graphdb.GraphDBClient with
function get_table as well as a set of predefined CIM queries.
"""

import re
from datetime import datetime
from typing import Dict, Iterable, List, Optional, Tuple, TypeVar, Union

import numpy as np
import pandas as pd

from cimsparql import queries
from cimsparql import query_support as sup
from cimsparql import ssh_queries, sv_queries, tp_queries
from cimsparql.constants import generating_types, impedance_variables, mrid_variable, ratings
from cimsparql.type_mapper import TypeMapper
from cimsparql.url import Prefix

CimModelType = TypeVar("CimModelType", bound="CimModel")


class CimModel(Prefix):
    """Used to query with sparql queries (typically CIM)"""

    def __init__(self, mapper: TypeMapper, *args, **kwargs) -> None:
        self._setup_client(*args, **kwargs)
        self.mapper = mapper

    @property
    def date_version(self) -> datetime:
        """Activation date for this repository"""
        try:
            date_version = self._date_version
        except AttributeError:
            repository_date = self._get_table_and_convert(queries.version_date())
            date_version = repository_date["activationDate"].values[0]
            if isinstance(date_version, np.datetime64):
                date_version = self._date_version = date_version.astype("<M8[s]").astype(datetime)
        return date_version

    @property
    def mapper(self) -> TypeMapper:
        """Mapper used to convert str → type described by ontology in Graphdb

        Getter:
           Returns a mapper that can be used by self or other instances
        Setter:
           Sets mapper for self. Query GraphDB if not provided (None)
        """
        return self._mapper

    @mapper.setter
    def mapper(self, mapper: Optional[TypeMapper] = None) -> None:
        if mapper is None and "rdfs" in self.prefixes:
            self._mapper = TypeMapper(self)
        else:
            self._mapper = mapper

    def _query_with_header(self, query: str, limit: Optional[int] = None) -> str:
        query = "\n".join([self.header_str(query), query])
        if limit is not None:
            query += f" limit {limit}"
        return query

    def bus_data(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        limit: Optional[int] = None,
        dry_run: bool = False,
        mrid: str = mrid_variable,
        name: str = "?name",
    ) -> pd.DataFrame:
        """Query name of topological nodes (TP query).

        Args:
           region: Limit to region (use None to get all)
           sub_region: True - assume sub regions, False - assume region
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = queries.bus_data(region, sub_region, mrid, name)
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index=mrid[1:])

    def phase_tap_changers(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        with_tap_changer_values: bool = True,
        impedance: Iterable[str] = impedance_variables,
        tap_changer_objects: Iterable[str] = ("high", "low", "neutral"),
        mrid: str = mrid_variable,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Get list of phase tap changers"""
        query = queries.phase_tap_changer_query(
            region, sub_region, with_tap_changer_values, impedance, tap_changer_objects, mrid
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index=mrid[1:])

    def loads(
        self,
        load_type: List[str],
        load_vars: Optional[Tuple[str]] = None,
        region: Union[str, List[str]] = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        station_group_optional: bool = True,
        station_group: bool = False,
        with_sequence_number: bool = False,
        network_analysis: bool = True,
        mrid: str = mrid_variable,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query load data

        Args:
           load_type: List of load types. Allowed: "ConformLoad", "NonConformLoad", "EnergyConsumer"
           load_vars: List of additional load vars to return. Possible are 'p' and/or 'q'.
           region: Limit to region
           sub_region: Assume region is a sub_region
           connectivity: Include connectivity mrids
           station_group: return station group mrid (if any)
           with_sequence_number: Include the sequence numbers in output
           network_analysis: Include only network analysis enabled components
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Returns:
           DataFrame: with mrid as index and columns ['terminal_mrid', 'bid_marked_code', 'p', 'q',
           'station_group']

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.loads(load_type=['ConformLoad', 'NonConformLoad'])
        """
        query = queries.load_query(
            load_type,
            load_vars,
            region,
            sub_region,
            connectivity,
            station_group_optional,
            with_sequence_number,
            network_analysis,
            station_group,
            self.cim_version,
            mrid,
        )
        if dry_run:
            return self._query_with_header(query, limit)

        columns = {} if load_vars is None else {var: float for var in load_vars}
        return self._get_table_and_convert(query, limit, index=mrid[1:], columns=columns)

    def wind_generating_units(
        self,
        limit: Optional[int] = None,
        network_analysis: bool = True,
        mrid: str = mrid_variable,
        name: str = "?name",
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query wind generating units

        Args:
           network_analysis: Include only network analysis enabled components
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Returns:.
           wind_generating_units: with mrid as index and columns ['station_group', 'market_code',
           'maxP', 'allocationMax', 'allocationWeight', 'minP', 'name', 'power_plant_mrid']

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.wind_generating_units(limit=10)

        """
        query = queries.wind_generating_unit_query(network_analysis, mrid, name)
        if dry_run:
            return self._query_with_header(query, limit)
        float_list = ["maxP", "allocationMax", "allocationWeight", "minP"]
        columns = {var: float for var in float_list}
        return self._get_table_and_convert(query, limit, index=mrid[1:], columns=columns)

    def synchronous_machines(
        self,
        sync_vars: Tuple[str] = ("sn", "p", "q"),
        region: Union[str, List[str]] = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        station_group_optional: bool = True,
        with_sequence_number: bool = False,
        network_analysis: bool = True,
        u_groups: bool = False,
        mrid: str = mrid_variable,
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query synchronous machines

        Args:
           synchronous_vars: additional vars to include in output
           region: Limit to region
           sub_region: Assume region is a sub_region
           connectivity: Include connectivity mrids
           station_group_optional: Assume station group is optional
           with_sequence_number: add this numbers
           network_analysis: query SN:Equipment.networkAnalysisEnable
           u_groups: Filter out station groups where name starts with 'U-'
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.synchronous_machines(limit=10)
        """
        terminal_mrid: str = "?t_mrid"

        query = queries.synchronous_machines_query(
            sync_vars,
            region,
            sub_region,
            connectivity,
            station_group_optional,
            self.cim_version,
            with_sequence_number,
            network_analysis,
            u_groups,
            terminal_mrid,
            mrid,
            name,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        float_list = ["maxP", "minP", "allocationMax", "allocationWeight", *sync_vars]
        columns = {var: float for var in float_list}
        return self._get_table_and_convert(query, limit, index=mrid[1:], columns=columns)

    def connections(
        self,
        rdf_types: Union[str, Tuple[str]] = ("cim:Breaker", "cim:Disconnector"),
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        mrid: str = mrid_variable,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query connectors

        Args:
           rdf_types: Only cim:breaker and cim:Disconnector allowed
        Returns:

        Example:
           region: Limit to region
           sub_region: Assume region is a sub_region
           limit: return first 'limit' number of rows
           connectivity: Include connectivity mrids
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.connections(limit=10)
        """
        query = queries.connection_query(
            self.cim_version, rdf_types, region, sub_region, connectivity, mrid
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index=mrid[1:])

    def borders(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        ignore_hvdc: bool = True,
        with_market_code: bool = False,
        market_optional: bool = False,
        mrid: str = mrid_variable,
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Retrieve ACLineSegments where one terminal is inside and the other is outside the region

        Args:
            region: Inside area
            sub_region: regions is sub areas
            ignore_hvdc: ignore ac lines with HVDC in name
            with_marked_code: include SN:Line.marketCode
            market_optional: include lines without market code if with_marked_code set
            limit: return first 'limit' number of rows
            dry_run: return string with sql query

        """
        query = queries.borders_query(
            self.cim_version,
            region,
            sub_region,
            ignore_hvdc,
            with_market_code,
            market_optional,
            mrid,
            name,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index=mrid[1:])

    def converters(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        converter_types: Iterable[str] = ("VoltageSource", "DC"),
        mrid: str = mrid_variable,
        name: str = "?name",
        sequence_numbers: Optional[List[int]] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        query = queries.converters(
            region, sub_region, converter_types, mrid, name, sequence_numbers
        )
        if dry_run:
            return self._query_with_header(query)
        return self._get_table_and_convert(query, index=mrid[1:])

    def transformers_connected_to_converter(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        converter_types: Iterable[str] = ("VoltageSource", "DC"),
        mrid: str = mrid_variable,
        name: str = "?name",
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query list of transformer connected at a converter (Voltage source or DC)

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           converter_types: VoltageSource or DC (ALG:VoltageSourceConverter or ALG:DCConverter)
           dry_run: return string with sql query

        """
        query = queries.transformers_connected_to_converter(
            region, sub_region, converter_types, mrid, name
        )
        if dry_run:
            return self._query_with_header(query)
        return self._get_table_and_convert(query, index=mrid[1:])

    def ac_lines(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        limit: Optional[int] = None,
        connectivity: Optional[str] = None,
        rates: Tuple[str] = ratings,
        network_analysis: bool = True,
        with_market: bool = False,
        temperatures: Optional[List[int]] = None,
        impedance: Iterable[str] = impedance_variables,
        mrid: str = mrid_variable,
        name: str = "?name",
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query ac line segments

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           limit: return first 'limit' number of rows
           connectivity: Include connectivity mrids as column name '{connectivity}_{1|2}'
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           with_market: include marked connections in output
           temperatures: include temperature correction factors in output
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.ac_lines(limit=10)
        """
        query = queries.ac_line_query(
            self.cim_version,
            self.ns["cim"],
            region,
            sub_region,
            connectivity,
            rates,
            network_analysis,
            with_market,
            temperatures,
            impedance,
            mrid,
            name,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        ac_lines = self._get_table_and_convert(query, limit=limit)
        if temperatures is not None:
            for temperature in temperatures:
                column = f"{sup.negpos(temperature)}_{abs(temperature)}_factor"
                ac_lines.loc[ac_lines[column].isna(), column] = 1.0
        return ac_lines

    def series_compensators(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        network_analysis: bool = True,
        with_market: bool = False,
        mrid: str = mrid_variable,
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query series compensators

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           limit: return first 'limit' number of rows
           connectivity: Include connectivity mrids
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.series_compensators(limit=10)
        """
        query = queries.series_compensator_query(
            self.cim_version,
            region,
            sub_region,
            connectivity,
            network_analysis,
            with_market,
            mrid,
            name,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def transformers(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        rates: Tuple[str] = ratings,
        network_analysis: bool = True,
        with_market: bool = False,
        impedance: Iterable[str] = impedance_variables,
        mrid: str = "?p_mrid",
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query transformer windings

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           connectivity: Include connectivity mrids
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           impedance: values returned
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.transformers(limit=10)
        """
        query = queries.transformer_query(
            region,
            sub_region,
            connectivity,
            rates,
            network_analysis,
            with_market,
            mrid,
            name,
            impedance,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def two_winding_transformers(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        rates: Tuple[str] = ratings,
        network_analysis: bool = True,
        with_market: bool = False,
        impedance: Iterable[str] = impedance_variables,
        mrid: str = "?p_mrid",
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query two-winding transformer

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           impedance: values returned
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.two_winding_transformers(limit=10)
        """
        query = queries.two_winding_transformer_query(
            region, sub_region, rates, network_analysis, with_market, mrid, name, impedance
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def three_winding_transformers(
        self,
        region: Optional[Union[str, List[str]]] = None,
        sub_region: bool = False,
        rates: Tuple[str] = ratings,
        network_analysis: bool = True,
        with_market: bool = False,
        impedance: Iterable[str] = impedance_variables,
        mrid: str = "?p_mrid",
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query three-winding transformer. Return as three two-winding transformers.

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_regionb
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           impedance: values returned
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.two_winding_transformers(limit=10)
        """
        query = queries.three_winding_transformer_query(
            region, sub_region, rates, network_analysis, with_market, mrid, name, impedance
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def disconnected(
        self, index: Optional[str] = None, limit: Optional[int] = None, dry_run: bool = False
    ) -> pd.DataFrame:
        """Query disconneced status from ssh profile (not available in GraphDB)

        Args:
           index: Column to use as index
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = ssh_queries.disconnected(self.cim_version)
        if dry_run:
            return self._query_with_header(query, limit)
        return self.get_table(query, index=index, limit=limit)

    def ssh_synchronous_machines(
        self, limit: Optional[int] = None, dry_run: bool = False
    ) -> pd.DataFrame:
        """Query synchronous machines from ssh profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = ssh_queries.synchronous_machines()
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {"p": float, "q": float, "controlEnabled": bool}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    def ssh_load(
        self,
        rdf_types: Optional[List[str]] = None,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query load data from ssh profile (not available in GraphDB)

        Args:
           rdf_types: allowed ["cim:ConformLoad", "cim:NonConformLoad"]
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        """
        if rdf_types is None:
            rdf_types = ["cim:ConformLoad", "cim:NonConformLoad"]
        query = ssh_queries.load(rdf_types)
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {"p": float, "q": float}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    def ssh_generating_unit(
        self, rdf_types: List[str] = None, limit: Optional[int] = None, dry_run: bool = False
    ) -> pd.DataFrame:
        """Query generating units from ssh profile (not available in GraphDB)

        Args:
           rdf_types: allowed
               ["cim:HydroGeneratingUnit", "cim:ThermalGeneratingUnit", "cim:WindGeneratingUnit"]
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        """
        if rdf_types is None:
            rdf_types = [f"cim:{unit}GeneratingUnit" for unit in generating_types]
        query = ssh_queries.generating_unit(rdf_types)
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {"normalPF": float}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    def terminal(self, limit: Optional[int] = None, dry_run: bool = False) -> pd.DataFrame:
        """Query terminals from tp profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = tp_queries.terminal(self.cim_version)
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {"connected": bool}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    def topological_node(self, limit: Optional[int] = None, dry_run: bool = False) -> pd.DataFrame:
        """Query topological nodes from tp profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = tp_queries.topological_node()
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {"BaseVoltage": float}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    def powerflow(
        self, power: Tuple[str] = ("p", "q"), limit: Optional[int] = None, dry_run: bool = False
    ) -> pd.DataFrame:
        """Query powerflow from sv profile (not available in GraphDB)

        Args:
           power: Allowed ['p','q']
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = sv_queries.powerflow(power)
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {x: float for x in power}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    def voltage(
        self,
        voltage_vars: Iterable[str] = ("v", "angle"),
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """Query voltage from sv profile (not available in GraphDB)

        Args:
           voltage_vars: allowed ["v", "angle"]
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = sv_queries.voltage(voltage_vars)
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {x: float for x in voltage_vars}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    def tapstep(self, limit: Optional[int] = None, dry_run: bool = False) -> pd.DataFrame:
        """Query tapstep from sv profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = sv_queries.tapstep()
        if dry_run:
            return self._query_with_header(query, limit)
        columns = {"position": float}
        return self._get_table_and_convert(query, limit, index=mrid_variable[1:], columns=columns)

    @property
    def regions(self) -> pd.DataFrame:
        """Query regions

        Property:
           regions: List of regions in database

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.regions
        """
        query = queries.regions_query(mrid_variable)
        return self._get_table_and_convert(query, limit=None, index=mrid_variable[1:])

    @property
    def empty(self) -> bool:
        """Identify empty GraphDB repo"""
        try:
            self._get_table("SELECT * \n WHERE { ?s ?p ?o }", limit=1)
            return False
        except IndexError:
            return True

    @staticmethod
    def _assign_column_types(
        result: pd.DataFrame, columns: Dict[str, Union[bool, str, float, int]]
    ) -> None:
        for column, column_type in columns.items():
            if column_type is str:
                continue
            if column_type is bool:
                result[column] = result[column].str.contains("true", flags=re.IGNORECASE)
            else:
                result.loc[result[column] == "None", column] = ""
                result[column] = pd.to_numeric(result[column], errors="coerce").astype(column_type)

    @classmethod
    def col_map(cls, data_row, columns) -> Tuple[Dict[str, str]]:
        columns = {} if columns is None else columns
        col_map = cls._col_map(data_row)
        return col_map, {col: columns[col] for col in set(columns).difference(col_map)}

    def get_table(
        self,
        query: str,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        map_data_types: bool = True,
        custom_maps: Optional[Dict] = None,
        columns: Optional[Dict] = None,
    ) -> pd.DataFrame:
        """Gets given table from the configured database.

        Args:
           query: to sparql server
           index: column name to use as index
           limit: limit number of resulting rows
           map_data_types: gets datatypes from the configured graphdb & maps the types in the result
                  to correct python types
           custom_maps: dictionary of 'sparql_datatype': function to apply on columns with that
                  type. Overwrites sparql map for the types specified.
           columns: dictionary of 'column_name': function,
                  uses pandas astype on the column, or applies function.
                  Sparql map overwrites columns when available

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> query = 'select * where { ?subject ?predicate ?object }'
           >>> gdbc.get_table(query, limit=10)
        """
        try:
            result, data_row = self._get_table(query, limit)
        except IndexError:
            return pd.DataFrame([])

        if map_data_types and self.mapper is not None:
            col_map, columns = self.col_map(data_row, columns)
            result = self.mapper.map_data_types(result, col_map, custom_maps, columns)

        if index and not result.empty:
            result.set_index(index, inplace=True)
        return result

    @property
    def map_data_types(self) -> bool:
        try:
            return self.mapper.have_cim_version(self.prefixes["cim"])
        except AttributeError:
            return False

    @classmethod
    def _manual_convert_types(
        cls: CimModelType, df: pd.DataFrame, columns: Dict, index: str
    ) -> pd.DataFrame:
        if columns is None:
            columns = {}
        reset_index = index is not None
        if reset_index:
            df.reset_index(inplace=True)
        df = df.astype(str)
        cls._assign_column_types(df, columns)
        if reset_index:
            df.set_index(index, inplace=True)
        return df

    def _get_table_and_convert(
        self,
        query: str,
        limit: Optional[int] = None,
        index: Optional[str] = None,
        custom_maps: Optional[Dict] = None,
        columns: Optional[Dict] = None,
    ) -> pd.DataFrame:
        result = self.get_table(
            query, index, limit, map_data_types=True, custom_maps=custom_maps, columns=columns
        )

        if not self.map_data_types and len(result) > 0:
            result = self._manual_convert_types(result, columns, index)

        return result



from __future__ import annotations

import re
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional

import pandas as pd
from dateutil import parser

from cimsparql.query_support import combine_statements, unionize

if TYPE_CHECKING:
    from cimsparql.model import CimModel

as_type_able = [int, float, str, "Int64", "Int32", "Int16"]

python_type_map = {
    "string": str,
    "integer": int,
    "boolean": lambda x: x.lower() == "true",
    "float": float,
    "dateTime": parser.parse,
}

uri_snmst = re.compile("[^\\#]*(.\\#\\_)")
sparql_type_map = {"literal": str, "uri": lambda x: uri_snmst.sub("", x) if x is not None else ""}


class TypeMapperQueries:
    @property
    def generals(self) -> List[List[str]]:
        """For sparql-types that are not sourced from objects of type rdf:property, sparql & type are
        required

        Sparql values should be like: http://iec.ch/TC57/2010/CIM-schema-cim15#PerCent this is how
        type or DataType usually looks like for each data point in the converted query result from
        SPARQLWrapper.

        type can be anything as long as it is represented in the python_type_map.
        """
        return [
            [
                "?sparql_type rdf:type rdfs:Datatype",
                "?sparql_type owl:equivalentClass ?range",
                'BIND(STRBEFORE(str(?range), "#") as ?prefix)',
                'BIND(STRAFTER(str(?range), "#") as ?type)',
            ]
        ]

    @property
    def prefix_general(self) -> List[str]:
        """Common query used as a base for all prefix_based queries."""
        return [
            "?sparql_type rdf:type rdf:Property",
            "?sparql_type rdfs:range ?range",
            'BIND(STRBEFORE(str(?range), "#") as ?prefix)',
        ]

    @property
    def prefix_based(self) -> Dict[str, List[str]]:
        """Each prefix can have different locations of where DataTypes are described.

        Based on a object of type rdf:property & its rdfs:range, one has edit the query such that
        one ends up with the DataType.

        """
        return {
            "https://www.w3.org/2001/XMLSchema": ["?range rdfs:label ?type"],
            "https://iec.ch/TC57/2010/CIM-schema-cim15": [
                "?range owl:equivalentClass ?class",
                "?class rdfs:label ?type",
            ],
        }

    @property
    def query(self) -> str:
        select_query = "SELECT ?sparql_type ?type ?prefix"

        grouped_generals = [combine_statements(*g, split=" .\n") for g in self.generals]
        grouped_prefixes = [
            combine_statements(*v, f'FILTER (?prefix = "{k}")', split=" .\n")
            for k, v in self.prefix_based.items()
        ]
        grouped_prefix_general = combine_statements(*self.prefix_general, split=" .\n")
        unionized_generals = unionize(*grouped_generals)
        unionized_prefixes = unionize(*grouped_prefixes)

        full_prefixes = combine_statements(grouped_prefix_general, unionized_prefixes, group=True)
        full_union = unionize(unionized_generals, full_prefixes, group=False)
        return f"{select_query}\nWHERE\n{{\n{full_union}\n}}"


class TypeMapper(TypeMapperQueries):
    def __init__(self, client: CimModel, custom_additions: Optional[Dict[str, Any]] = None) -> None:
        self.prefixes = client.prefixes
        custom_additions = custom_additions if custom_additions is not None else {}
        self.map = {**sparql_type_map, **self.get_map(client), **custom_additions}

    def have_cim_version(self, cim) -> bool:
        return cim in (val.split("#")[0] for val in self.map.keys())

    @staticmethod
    def type_map(df: pd.DataFrame) -> Dict[str, Any]:
        df["type"] = df["type"].str.lower()
        d = df.set_index("sparql_type").to_dict("index")
        return {k: python_type_map.get(v.get("type", "string")) for k, v in d.items()}

    @staticmethod
    def prefix_map(df: pd.DataFrame) -> Dict[str, Any]:
        df = df.loc[~df["prefix"].isna()].head()
        df["comb"] = df["prefix"] + "#" + df["type"]
        df = df.drop_duplicates("comb")
        d2 = df.set_index("comb").to_dict("index")
        return {k: python_type_map.get(v.get("type", "string")) for k, v in d2.items()}

    def get_map(self, client: CimModel) -> Dict[str, Any]:
        """Reads all metadata from the sparql backend & creates a sparql-type -> python type map

        Args:
            client: initialized CimModel

        Returns:
            sparql-type -> python type map

        """
        df = client.get_table(self.query, map_data_types=False)
        if df.empty:
            return {}
        type_map = self.type_map(df)
        prefix_map = self.prefix_map(df)
        xsd_map = {
            f"{self.prefixes['xsd']}#{xsd_type}": xsd_map
            for xsd_type, xsd_map in python_type_map.items()
        }
        return {**type_map, **prefix_map, **xsd_map}

    def get_type(
        self,
        sparql_type: str,
        missing_return: str = "identity",
        custom_maps: Optional[Dict[str, Any]] = None,
    ):
        """Gets the python type/function to apply on columns of the sparql_type

        Args:
            sparql_type:
            missing_return: returns the identity-function if python- type/function is not found,
                else returns None

            custom_maps: dictionary on the form {'sparql_data_type': function/datatype} overwrites
                the default types gained from the graphdb. Applies the function/datatype on all
                columns in the DataFrame that are of the sparql_data_type

        Returns:
            python datatype or function to apply on DataFrame columns

        """
        type_map = {**self.map, **custom_maps} if custom_maps is not None else self.map
        try:
            return type_map[sparql_type]
        except KeyError:
            warnings.warn(f"{sparql_type} not found in the sparql -> python type map")
            if missing_return == "identity":
                return lambda x: x
            return None

    def convert_dict(
        self, d: Dict, drop_missing: bool = True, custom_maps: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """Converts a col_name -> sparql_datatype map to a col_name -> python_type map

        Args:
            d: dictionary with {'column_name': 'sparql type/DataType'}
            drop_missing: drops columns where no corresponding python type could be found
            custom_maps: dictionary on the form {'sparql_data_type': function/datatype} overwrites
                the default types gained from the graphdb. Applies the function/datatype on all
                columns in the DataFrame that are of the sparql_data_type.

        Returns:
            col_name -> python_type/function map

        """
        missing_return = "None" if drop_missing else "identity"
        base = {
            column: self.get_type(data_type, missing_return, custom_maps)
            for column, data_type in d.items()
        }
        if drop_missing:
            return {key: value for key, value in base.items() if value is not None}
        return base

    @staticmethod
    def map_base_types(df: pd.DataFrame, type_map: Dict) -> pd.DataFrame:
        """Maps the datatypes in type_map which can be used with the df.astype function

        Args:
            df:
            type_map: {'column_name': type/function} map of functions/types to apply on the columns

        Returns:
            mapped DataFrame

        """
        as_type_able_columns = {c for c, datatype in type_map.items() if datatype in as_type_able}
        if not df.empty:
            df = df.astype({column: type_map[column] for column in as_type_able_columns})
        return df

    @staticmethod
    def map_exceptions(df: pd.DataFrame, type_map: Dict) -> pd.DataFrame:
        """Maps the functions/datatypes in type_map which cant be done with the df.astype function

        Args:
            df:
            type_map: {'column_name': type/function} map of functions/types to apply on the columns

        Returns:
            mapped DataFrame

        """
        ex_columns = {c for c, datatype in type_map.items() if datatype not in as_type_able}
        for column in ex_columns:
            df[column] = df[column].apply(type_map[column])
        return df

    def map_data_types(
        self, df: pd.DataFrame, col_map: Dict, custom_maps: Dict = None, columns: Dict = None
    ) -> pd.DataFrame:
        """Maps the dtypes of a DataFrame to the python-corresponding types of the sparql-types from the
        source data

        Args:
            df: DataFrame with columns to be converted
            data_row: a complete row with data from the source data of which the DataFrame is
                constructed from
            custom_maps: dictionary on the form {'sparql_data_type': function/datatype} overwrites
                the default types gained from the graphdb. Applies the function/datatype on all
                columns in the DataFrame that are of the sparql_data_type.
            columns: dictionary on the form {'DataFrame_column_name: function/datatype} overwrites
                the default types gained from the graphdb.  Applies the function/datatype on the
                column.

        Returns:
            mapped DataFrame

        """
        type_map = {**self.convert_dict(col_map, custom_maps=custom_maps), **columns}
        df = self.map_base_types(df, type_map)
        df = self.map_exceptions(df, type_map)
        return df



from typing import Dict, Iterable, List, Optional, Union

from cimsparql.cim import (
    CNODE_CONTAINER,
    DELIVERYPOINT,
    EQUIP_CONTAINER,
    GEO_REG,
    ID_OBJ,
    SUBSTATION,
)
from cimsparql.constants import con_mrid_str, sequence_numbers


def base_voltage(mrid: str, var: str) -> List[str]:
    return [
        f"{mrid} cim:ConductingEquipment.BaseVoltage ?obase",
        f"?obase cim:BaseVoltage.nominalVoltage {var}",
    ]


def terminal_sequence_query(cim_version: int, var: str, t_mrid: str = "?t_mrid") -> List[str]:
    query_list = []
    for nr in sequence_numbers:
        t_sequence_mrid = f"{t_mrid}_{nr}"
        query_list.extend(
            [
                rdf_type_tripler(t_sequence_mrid, "cim:Terminal"),
                f"{t_sequence_mrid} cim:Terminal.ConductingEquipment ?mrid",
                f"{t_sequence_mrid} cim:{acdc_terminal(cim_version)}.sequenceNumber {nr}",
            ]
        )
        if var is not None:
            query_list.append(f"{t_sequence_mrid} cim:Terminal.ConnectivityNode ?{var}_{nr}")
    return query_list


def operational_limit(mrid: str, rate: str, limitset: str = "oplimset") -> List[str]:
    return [
        f"?{limitset} cim:OperationalLimitSet.Equipment {mrid}",
        f"?p_lim{rate} cim:OperationalLimit.OperationalLimitSet ?{limitset}",
        rdf_type_tripler(f"?p_lim{rate}", "cim:ActivePowerLimit"),
        f"?p_lim{rate} {ID_OBJ}.name '{rate}@20'",
        f"?p_lim{rate} cim:ActivePowerLimit.value ?rate{rate}",
    ]


def region_name_query(
    region: str, sub_region: bool, sub_geographical_region: str, region_var: str = "?region"
) -> List[str]:
    if sub_region:
        return [f"{sub_geographical_region} SN:IdentifiedObject.shortName {region}"]
    return [
        f"{sub_geographical_region} {GEO_REG}.Region {region_var}",
        f"{region_var} {ID_OBJ}.name {region}",
    ]


def region_query(
    region: Union[str, List[str]], sub_region: bool, container: str, sub_geographical_region: str
) -> List[str]:
    if region is None:
        return []
    query = [f"?{container} cim:{container}.Region {sub_geographical_region}"]
    if isinstance(region, str):
        query.extend(region_name_query(f"'{region}'", sub_region, sub_geographical_region))
    elif isinstance(region, list):
        query.extend(region_name_query("?r_na", sub_region, sub_geographical_region))
        query.append("FILTER regex(?r_na, '" + "|".join(region) + "')")
    else:
        raise NotImplementedError("region must be either str or List")
    return query


def sequence_variables(var: str = con_mrid_str) -> List[str]:
    return [f"?{var}_{nr}" for nr in sequence_numbers]


def _xsd_type(cim: str, var: str) -> str:
    return f"^^<{cim}{var}>"


def acdc_terminal(cim_version: int) -> str:
    return "ACDCTerminal" if cim_version > 15 else "Terminal"


def predicate_list(subject: str, predicate: str, properties: Dict[str, str]) -> List[str]:
    return [f"{subject} {predicate}.{property} {object}" for property, object in properties.items()]


def rdf_type_tripler(subject: str, predicate: str) -> str:
    return f"{subject} rdf:type {predicate}"


def include_market(with_market: bool, variables: List[str], where_list: List[str]) -> None:
    if with_market:
        variables.extend(sequence_variables("bidzone"))
        where_list.extend([market_code_query(terminal_nr) for terminal_nr in sequence_numbers])


def market_code_query(nr: int = None):
    nr_s = "" if nr is None else f"_{nr}"
    return group_query(
        [
            f"?t_mrid{nr_s} cim:Terminal.ConnectivityNode ?con{nr_s}",
            f"?con{nr_s} {CNODE_CONTAINER} ?container{nr_s}",
            f"?container{nr_s} {SUBSTATION} ?substation{nr_s}",
            f"?substation{nr_s} {DELIVERYPOINT} ?m_d_p{nr_s}",
            f"?m_d_p{nr_s} SN:MarketDeliveryPoint.BiddingArea ?barea{nr_s}",
            f"?barea{nr_s} SN:BiddingArea.marketCode ?bidzone{nr_s}",
        ],
        command="OPTIONAL",
    )


def terminal_where_query(
    cim_version: int = 15,
    var: str = con_mrid_str,
    with_sequence_number: bool = False,
    terminal_mrid: str = "?t_mrid",
) -> List[str]:
    out = [
        rdf_type_tripler(terminal_mrid, "cim:Terminal"),
        f"{terminal_mrid} cim:Terminal.ConductingEquipment ?mrid",
    ]
    if var is not None:
        out.append(f"{terminal_mrid} cim:Terminal.ConnectivityNode ?{var}")

    if with_sequence_number:
        out.append(
            f"{terminal_mrid} cim:{acdc_terminal(cim_version)}.sequenceNumber ?sequenceNumber"
        )
    return out


def _temperature_list(temperature: float, xsd: str, curve: str) -> List[str]:
    signed_temperature = f"{negpos(temperature)}_{abs(temperature)}"
    subject = f"?t{signed_temperature}"
    percent = f"?{signed_temperature}_factor"
    temperature_value = f"'{temperature:0.1f}'{xsd}"
    properties = {"Curve": curve, "temperature": temperature_value, "percent": percent}
    return predicate_list(subject, "ALG:TemperatureCurveData", properties)


def temp_correction_factors(
    mrid: str, cim: str, temperatures: Iterable, temperature_mrid: str = "?temp_mrid"
) -> List[str]:
    where_list = [
        rdf_type_tripler(temperature_mrid, "ALG:TemperatureCurveDependentLimit"),
        f"{temperature_mrid} ALG:LimitDependency.Equipment {mrid}",
        f"{temperature_mrid} ALG:TemperatureCurveDependentLimit.TemperatureCurve ?tcur",
    ]
    xsd = _xsd_type(cim, "Temperature")
    for temperature in temperatures:
        where_list.extend(_temperature_list(temperature, xsd, "?tcur"))
    return where_list


def bid_market_code_query() -> List[str]:
    return [
        f"?mrid {EQUIP_CONTAINER} ?eq_container",
        f"?eq_container {SUBSTATION} ?substation",
        f"?substation {DELIVERYPOINT} ?m_d_p",
        "?m_d_p SN:MarketDeliveryPoint.BiddingArea ?barea",
        "?barea SN:BiddingArea.marketCode ?bidzone",
    ]


def to_variables(vars: Iterable[str]) -> List[str]:
    return [f"?{var}" for var in vars]


def combine_statements(*args, group: bool = False, split: str = "\n") -> str:
    """Join *args

    Args:
       group: return enclosed by {...}
       split: join *args by this

    Example:
       >>> import os
       >>> where_list = ['?mrid rdf:type cim:ACLineSegment', '?mrid cim:ACLineSegment.r ?r']
       >>> combine_statements(where_list,group=True, split='\n')
    """
    return "{\n" + split.join(args) + "\n}" if group else split.join(args)


def negpos(val: Union[float, int]) -> str:
    """Convert 'sign' to text"""
    return "minus" if val < 0 else "plus"


def select_statement(variables: Optional[List[str]] = None) -> str:
    """Combine variables in an select statement"""
    variables = "*" if variables is None else " ".join(variables)
    return f"SELECT {variables}"


def group_query(
    x: List[str], command: str = "WHERE", split: str = ".\n", group: bool = True
) -> str:
    """Group Query

    Args:
       x: List of objects to group
       command: to operate on group
       split, group: (see: combine_statements)

    Example:
       >>> import os
       >>> where_list = ['?mrid rdf:type cim:ACLineSegment', '?mrid cim:ACLineSegment.r ?r']
       >>> group_query(where_list, group=True, split= '.\n')
    """
    return command + " " + combine_statements(*x, group=group, split=split)


def unionize(*args: str, group: bool = True):
    if group:
        args = tuple(f"{{\n{arg}\n}}" for arg in args)
    return "\nUNION\n".join(args)


def get_name(mrid: str, name: str, alias: bool = False) -> str:
    param = "aliasName" if alias else "name"
    return f"{mrid} {ID_OBJ}.{param} {name}"


def border_filter(region: Union[str, List[str]], area1: str, area2: str) -> List[str]:
    """Border filter where one area is in and the other is out"""

    def _in_first(var1: str, var2: str, regions: str) -> List[str]:
        """Return filter for inclusion of first variable and not second"""
        return [f"FILTER (regex({var1}, '{regions}'))", f"FILTER (!regex({var2}, '{regions}'))"]

    regions = "|".join(region) if isinstance(region, list) else region
    return [
        combine_statements(*_in_first(area1, area2, regions)),
        combine_statements(*_in_first(area2, area1, regions)),
    ]



from typing import Iterable, List, Tuple

# Allowed
allowed_load_types: Tuple[str] = ("ConformLoad", "NonConformLoad", "EnergyConsumer")

generating_types: Iterable[str] = ("Hydro", "Thermal", "Wind")

# Available ratings
ratings: Tuple[str] = ("Normal", "Warning", "Overload")

union_split: str = "\n} UNION\n{"
sequence_numbers: List[int] = [1, 2]

# Default variable names
con_mrid_str: str = "connectivity_mrid"
mrid_variable: str = "?mrid"

impedance_variables: Iterable[str] = ("r", "x")



from functools import reduce
from operator import iconcat
from typing import Iterable, List, Optional, Union

import cimsparql.query_support as sup
from cimsparql.cim import (
    ACLINE,
    CNODE_CONTAINER,
    EQUIP_CONTAINER,
    GEO_REG,
    ID_OBJ,
    SUBSTATION,
    SYNC_MACH,
    TR_WINDING,
)
from cimsparql.constants import allowed_load_types, mrid_variable, sequence_numbers, union_split
from cimsparql.transformer_windings import terminal, transformer_common


def version_date() -> str:
    name: str = "?name"
    variables = [mrid_variable, name, "?activationDate"]
    where_list = [
        sup.rdf_type_tripler("?marketDefinitionSet", "SN:MarketDefinitionSet"),
        f"?marketDefinitionSet {ID_OBJ}.mRID {mrid_variable}",
        sup.get_name("?marketDefinitionSet", name),
        "?marketDefinitionSet SN:MarketDefinitionSet.activationDate ?activationDate",
        f"FILTER regex({name}, 'ScheduleResource')",
    ]
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def regions_query(mrid: str) -> str:
    variables = [mrid, "?shortName"]
    region_variable = "?subgeoreg"
    where_list = [
        sup.rdf_type_tripler(mrid, GEO_REG),
        f"{mrid} SN:IdentifiedObject.shortName ?shortName",
        f"{mrid} {GEO_REG}.Region {region_variable}",
    ]
    names = {mrid: ["?name", "?alias_name"], region_variable: ["?region", "?region_name"]}
    for name_mrid, (name, alias_name) in names.items():
        where_list.append(sup.get_name(name_mrid, name))
        where_list.append(sup.get_name(name_mrid, alias_name, alias=True))
        variables.extend([name, alias_name])
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def phase_tap_changer_query(
    region: Union[str, List[str]],
    sub_region: bool,
    with_tap_changer_values: bool,
    impedance: Iterable[str],
    tap_changer_objects: Iterable[str],
    mrid: str,
) -> str:
    variables = [mrid, "?w_mrid_1", "?w_mrid_2", "?t_mrid_1", "?t_mrid_2"]
    tap = "?tap"
    where_list = [
        sup.rdf_type_tripler(mrid, TR_WINDING),
        f"{mrid} cim:TransformerEnd.PhaseTapChanger {tap}",
        f"{mrid} {TR_WINDING}.PowerTransformer ?pt",
    ]

    if with_tap_changer_values:
        variables.extend([tap, "?phase_incr"] + sup.to_variables(tap_changer_objects))
        properties = {f"{obj}Step": f"?{obj}" for obj in tap_changer_objects}
        where_list.extend(
            [
                *sup.predicate_list(tap, "cim:TapChanger", properties),
                "?tap cim:PhaseTapChangerLinear.stepPhaseShiftIncrement ?phase_incr",
            ]
        )

    if impedance is not None:
        variables.extend(sup.to_variables(impedance))
        where_list.extend([f"?w_mrid_1 {TR_WINDING}.{imp} ?{imp}" for imp in impedance])

    if region is not None:
        where_list.extend([f"?pt {EQUIP_CONTAINER} ?Substation"])
        where_list.extend(sup.region_query(region, sub_region, "Substation", "?subgeoreg"))

    for i in sequence_numbers:
        where_list.extend(
            [
                sup.rdf_type_tripler(f"?term_{i}", "cim:Terminal"),
                f"?w_mrid_{i} {TR_WINDING}.PowerTransformer ?pt",
                f"?w_mrid_{i} cim:TransformerEnd.Terminal ?term_{i}",
                f'?term_{i} cim:Terminal.sequenceNumber "{i}"^^xsd:integer',
                f"?term_{i} {ID_OBJ}.mRID ?t_mrid_{i}",
            ]
        )

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def connectivity_names(mrid: str, name: str = "?name") -> str:
    variables = [mrid, name]
    where_list = [
        sup.rdf_type_tripler(mrid, "cim:ConnectivityNode"),
        sup.get_name(mrid, name),
    ]
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def bus_data(region: Union[str, List[str]], sub_region: bool, mrid: str, name: str) -> str:
    variables = [mrid, name]
    where_list = [sup.rdf_type_tripler(mrid, "cim:TopologicalNode"), sup.get_name(mrid, name)]

    if region is not None:
        where_list.extend(
            [
                f"{mrid} cim:TopologicalNode.ConnectivityNodeContainer ?cont",
                f"?cont {SUBSTATION} ?Substation",
                *sup.region_query(region, sub_region, "Substation", "?subgeoreg"),
            ]
        )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def load_query(
    load_type: List[str],
    load_vars: Iterable[str],
    region: Union[str, List[str]],
    sub_region: bool,
    connectivity: str,
    station_group_optional: bool,
    with_sequence_number: bool,
    network_analysis: bool,
    station_group: bool,
    cim_version: int,
    mrid: str,
) -> str:

    if not set(load_type).issubset(allowed_load_types) or not load_type:
        raise ValueError(f"load_type should be any combination of {allowed_load_types}")

    variables = [mrid, "?t_mrid", "?bidzone"]

    if with_sequence_number:
        variables.append("?sequenceNumber")

    if connectivity is not None:
        variables.append(f"?{connectivity}")

    cim_types = [sup.rdf_type_tripler(mrid, f"cim:{cim_type}") for cim_type in load_type]

    where_list = [
        sup.combine_statements(*cim_types, group=len(cim_types) > 1, split=union_split),
        *sup.terminal_where_query(cim_version, connectivity, with_sequence_number),
        *sup.bid_market_code_query(),
    ]

    if load_vars is not None:
        variables.extend([f"?{load}" for load in load_vars])
        where_list.append(
            sup.group_query(
                [f"{mrid} cim:EnergyConsumer.{load} ?{load}" for load in load_vars],
                command="OPTIONAL",
            )
        )

    if station_group:
        variables.append("?station_group")
        station_group_list = [
            f"{mrid} cim:NonConformLoad.LoadGroup ?lg",
            "?lg SN:NonConformLoadGroup.ScheduleResource ?sc_res",
            "?sc_res SN:ScheduleResource.marketCode ?station_group",
        ]
        if station_group_optional:
            where_list.append(sup.group_query(station_group_list, command="OPTIONAL"))
        else:
            where_list.extend(station_group_list)

    if network_analysis is not None:
        where_list.append(f"{mrid} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if region is not None:
        where_list.extend(
            [
                f"{mrid} {EQUIP_CONTAINER} ?cont",
                f"?cont {SUBSTATION} ?Substation",
                *sup.region_query(region, sub_region, "Substation", "?subgeoreg"),
            ]
        )

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def synchronous_machines_query(
    sync_vars: Iterable[str],
    region: Union[str, List[str]],
    sub_region: bool,
    connectivity: str,
    station_group_optional: bool,
    cim_version: int,
    with_sequence_number: bool,
    network_analysis: bool,
    u_groups: bool,
    terminal_mrid: str,
    mrid: str,
    name: str,
) -> str:

    variables = [
        mrid,
        name,
        terminal_mrid,
        "?station_group",
        "?market_code",
        "?maxP",
        "?allocationMax",
        "?allocationWeight",
        "?minP",
        "?bidzone",
        *[f"?{var}" for var in sync_vars],
    ]

    if connectivity is not None:
        variables.append(f"?{connectivity}")

    if with_sequence_number:
        variables.append("?sequenceNumber")

    properties = {"sn": "ratedS", "p": "p", "q": "q"}
    where_list = [
        sup.rdf_type_tripler(mrid, SYNC_MACH),
        sup.get_name(mrid, name),
        *sup.bid_market_code_query(),
        *sup.terminal_where_query(cim_version, connectivity, with_sequence_number, terminal_mrid),
        *[f"{mrid} {SYNC_MACH}.{lim} ?{lim}" for lim in ["maxQ", "minQ"] if lim in variables],
        *[
            sup.group_query(
                [f"{mrid} cim:RotatingMachine.{properties[var]} ?{var}"], command="OPTIONAL"
            )
            for var in sync_vars
        ],
    ]

    if network_analysis is not None:
        where_list.append(f"{mrid} SN:Equipment.networkAnalysisEnable {network_analysis}")

    station_group = [
        f"{mrid} cim:SynchronousMachine.GeneratingUnit ?gu",
        "?gu SN:GeneratingUnit.marketCode ?market_code",
        "?gu cim:GeneratingUnit.maxOperatingP ?maxP",
        "?gu cim:GeneratingUnit.minOperatingP ?minP",
        "?gu SN:GeneratingUnit.groupAllocationMax ?allocationMax",
        "?gu SN:GeneratingUnit.groupAllocationWeight ?allocationWeight",
        "?gu SN:GeneratingUnit.ScheduleResource ?ScheduleResource",
        "?ScheduleResource SN:ScheduleResource.marketCode ?station_group",
        sup.get_name("?ScheduleResource", "?st_gr_n", alias=True),
    ]

    if station_group_optional:
        where_list.append(sup.group_query(station_group, command="OPTIONAL"))
    else:
        where_list.extend(station_group)

    if not u_groups:
        where_list.append("FILTER (!bound(?st_gr_n) || (!regex(?st_gr_n, 'U-')))")

    if region is not None:
        where_list.extend(
            [
                f"{mrid} {EQUIP_CONTAINER} ?cont",
                f"?cont {SUBSTATION} ?Substation",
                *sup.region_query(region, sub_region, "Substation", "?subgeoreg"),
            ]
        )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def wind_generating_unit_query(network_analysis: bool, mrid: str, name: str):
    variables = [
        mrid,
        "?station_group",
        "?market_code",
        "?maxP",
        "?allocationMax",
        "?allocationWeight",
        "?minP",
        name,
        "?plant_mrid",
    ]
    where_list = [
        sup.rdf_type_tripler(mrid, "cim:WindGeneratingUnit"),
        f"{mrid} cim:GeneratingUnit.maxOperatingP ?maxP",
        f"{mrid} SN:GeneratingUnit.marketCode ?market_code",
        f"{mrid} cim:GeneratingUnit.minOperatingP ?minP",
        sup.get_name(mrid, name),
        f"{mrid} SN:WindGeneratingUnit.WindPowerPlant ?plant_mrid",
        f"{mrid} SN:GeneratingUnit.groupAllocationMax ?allocationMax",
        f"{mrid} SN:GeneratingUnit.groupAllocationWeight ?allocationWeight",
        f"{mrid} SN:GeneratingUnit.ScheduleResource ?sr",
        "?sr SN:ScheduleResource.marketCode ?station_group",
    ]

    if network_analysis is not None:
        where_list.append(f"{mrid} SN:Equipment.networkAnalysisEnable {network_analysis}")

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def two_winding_transformer_query(
    region: Union[str, List[str]],
    sub_region: bool,
    rates: Iterable[str],
    network_analysis: bool,
    with_market: bool,
    mrid: str,
    name: str,
    impedance: Iterable[str],
) -> str:
    variables = ["?t_mrid_1", "?t_mrid_2"]
    where_list = [*terminal(mrid, 1), *terminal(mrid, 2)]
    transformer_common(
        2,
        mrid,
        name,
        impedance,
        variables,
        where_list,
        with_market,
        region,
        sub_region,
        rates,
        network_analysis,
    )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def three_winding_transformer_query(
    region: Union[str, List[str]],
    sub_region: bool,
    rates: Iterable[str],
    network_analysis: bool,
    with_market: bool,
    mrid: str,
    name: str,
    impedance: Iterable[str],
) -> str:
    variables = ["?t_mrid_1", f"({mrid} as ?t_mrid_2)"]
    where_list = [*terminal(mrid, 1, lock_end_number=False)]
    transformer_common(
        3,
        mrid,
        name,
        impedance,
        variables,
        where_list,
        with_market,
        region,
        sub_region,
        rates,
        network_analysis,
    )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def transformer_query(
    region: Union[str, List[str]],
    sub_region: bool,
    connectivity: str,
    rates: Iterable[str],
    network_analysis: bool,
    with_market: bool,
    mrid: str,
    name: str,
    impedance: Iterable[str],
) -> str:
    variables = [name, mrid, "?w_mrid", "?endNumber", "?un", "?t_mrid"]
    variables.extend(sup.to_variables(impedance))
    where_list = [
        sup.rdf_type_tripler(mrid, "cim:PowerTransformer"),
        f"?w_mrid {TR_WINDING}.PowerTransformer {mrid}",
        f"?w_mrid {TR_WINDING}.ratedU ?un",
        "?w_mrid cim:TransformerEnd.endNumber ?endNumber",
        "?w_mrid cim:TransformerEnd.Terminal ?t_mrid",
        f"?w_mrid {ID_OBJ}.name {name}",
        *sup.predicate_list("?w_mrid", TR_WINDING, {z: f"?{z}" for z in impedance}),
    ]
    if with_market:
        variables.append("?bidzone")
        where_list.append(sup.market_code_query())

    if network_analysis is not None:
        where_list.append(f"{mrid} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if connectivity is not None:
        variables.append(f"?{connectivity}")
        where_list.append(f"?t_mrid cim:Terminal.ConnectivityNode ?{connectivity}")

    if region is not None:
        where_list.append(f"{mrid} {EQUIP_CONTAINER} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation", "?subgeoreg"))

    if rates:
        where_rate = ["?oplimitset cim:OperationalLimitSet.Terminal ?t_mrid"]

        for rate in rates:
            variables.append(f"?rate{rate}")
            where_rate.extend(sup.operational_limit(mrid, rate, "oplimitset"))
        where_list.append(sup.group_query(where_rate, command="OPTIONAL"))
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def series_compensator_query(
    cim_version: int,
    region: Union[str, List[str]],
    sub_region: bool,
    connectivity: str,
    network_analysis: Optional[bool],
    with_market: bool,
    mrid: str,
    name: str,
) -> str:
    variables = [mrid, "?x", "?un", name] + sup.sequence_variables("t_mrid")
    if connectivity is not None:
        variables += sup.sequence_variables(connectivity)

    where_list = [
        *sup.terminal_sequence_query(cim_version=cim_version, var=connectivity),
        sup.rdf_type_tripler(mrid, "cim:SeriesCompensator"),
        f"{mrid} cim:SeriesCompensator.x ?x",
        f"{mrid} cim:ConductingEquipment.BaseVoltage ?obase",
        "?obase cim:BaseVoltage.nominalVoltage ?un",
        sup.get_name(mrid, name),
    ]
    sup.include_market(with_market, variables, where_list)

    if network_analysis is not None:
        where_list.append(f"{mrid} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if region is not None:
        where_list += [
            f"{mrid} {EQUIP_CONTAINER} ?EquipmentContainer",
            f"?EquipmentContainer {SUBSTATION} ?Substation",
            *sup.region_query(region, sub_region, "Substation", "?subgeoreg"),
        ]

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def transformers_connected_to_converter(
    region: str, sub_region: bool, converter_types: Iterable[str], mrid: str, name: str
) -> str:
    variables = [mrid, "?t_mrid", name]
    converters = [
        sup.rdf_type_tripler("?volt", f"ALG:{converter}Converter") for converter in converter_types
    ]
    where_list = [
        sup.rdf_type_tripler(mrid, "cim:PowerTransformer"),
        sup.get_name(mrid, name, alias=True),
        f"?t_mrid cim:Terminal.ConductingEquipment {mrid}",
        "?t_mrid cim:Terminal.ConnectivityNode ?con",
        "?t_tvolt cim:Terminal.ConductingEquipment ?volt",
        "?t_tvolt cim:Terminal.ConnectivityNode ?con",
        sup.combine_statements(*converters, group=len(converters) > 1, split=union_split),
    ]
    if region is not None:
        where_list.append(f"{mrid} {EQUIP_CONTAINER} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation", "?subgeoreg"))
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def converters(
    region: str,
    sub_region: bool,
    converter_types: Iterable[str],
    mrid: str,
    name: str,
    sequence_numbers: Optional[List[int]],
) -> str:
    variables = [mrid, name]
    converters = [
        sup.rdf_type_tripler(mrid, f"ALG:{converter}Converter") for converter in converter_types
    ]
    where_list = [
        sup.get_name(mrid, name, alias=True),
        sup.combine_statements(*converters, group=len(converters) > 1, split=union_split),
    ]
    if sequence_numbers is not None:
        for num in sequence_numbers:
            variables += [f"?t_mrid_{num}"]
            where_list += [
                f"?t_{num} cim:Terminal.ConductingEquipment {mrid}",
                f'?t_{num} cim:Terminal.sequenceNumber "{num}"^^xsd:integer',
                f"?t_{num} cim:IdentifiedObject.mRID ?t_mrid_{num}",
            ]
    if region is not None:
        vc = f"{mrid} {EQUIP_CONTAINER} ?cont.\n?cont {SUBSTATION} ?Substation."
        dc = f"{mrid} ALG:DCConverter.DCPole ?pole.\n?pole {EQUIP_CONTAINER} ?Substation."
        where_list.extend(
            [
                sup.combine_statements(vc, dc, group=True, split=union_split),
                *sup.region_query(region, sub_region, "Substation", "?subgeoreg"),
            ]
        )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def borders_query(
    cim_version: int,
    region: Union[str, List[str]],
    sub_region: bool,
    ignore_hvdc: bool,
    with_market_code: bool,
    market_optional: bool,
    mrid: str,
    name: str,
) -> str:
    areas = sup.sequence_variables("area")
    variables = [name, mrid, *sup.sequence_variables("t_mrid"), *areas]
    border_filter = sup.border_filter(region, *areas)
    where_list = [
        sup.get_name(mrid, name),
        sup.rdf_type_tripler(mrid, ACLINE),
        *sup.terminal_sequence_query(cim_version=cim_version, var="con"),
        sup.combine_statements(*border_filter, group=True, split=union_split),
    ]
    for nr in sequence_numbers:
        where_list.extend(
            [
                f"?con_{nr} {CNODE_CONTAINER} ?cont_{nr}",
                f"?cont_{nr} {SUBSTATION} ?subs_{nr}",
                f"?subs_{nr} cim:Substation.Region ?reg_{nr}",
                *sup.region_name_query(f"?area_{nr}", sub_region, f"?reg_{nr}", f"?sreg_{nr}"),
            ]
        )

    if with_market_code:
        variables.append("?market_code")
        where_market = [
            f"{mrid} {EQUIP_CONTAINER} ?line_cont",
            "?line_cont SN:Line.marketCode ?market_code",
        ]
        where_list.append(
            sup.group_query(where_market, command="OPTIONAL" if market_optional else "")
        )

    if ignore_hvdc:
        where_list.append(sup.combine_statements(f"FILTER (!regex({name}, 'HVDC'))", group=True))

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def ac_line_query(
    cim_version: int,
    cim: str,
    region: Union[str, List[str]],
    sub_region: bool,
    connectivity: str,
    rates: Iterable[str],
    network_analysis: bool,
    with_market: bool,
    temperatures: Optional[List[int]],
    impedance: Iterable[str],
    mrid: str,
    name: str,
) -> str:
    variables = [
        name,
        mrid,
        "?length",
        "?un",
        *sup.sequence_variables("t_mrid"),
        *sup.to_variables(impedance),
    ]

    acline_properties = {z: f"?{z}" for z in impedance}

    where_list = [
        sup.rdf_type_tripler(mrid, ACLINE),
        f"{mrid} cim:Conductor.length ?length",
        sup.get_name(mrid, name),
        *sup.base_voltage(mrid, "?un"),
        *sup.terminal_sequence_query(cim_version=cim_version, var=connectivity),
        *sup.predicate_list(mrid, ACLINE, acline_properties),
    ]

    if connectivity is not None:
        variables.extend(sup.sequence_variables(connectivity))

    sup.include_market(with_market, variables, where_list)

    if network_analysis is not None:
        where_list.append(f"{mrid} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if region is not None:
        where_list.extend(
            [
                f"{mrid} {EQUIP_CONTAINER} ?Line",
                *sup.region_query(region, sub_region, "Line", "?subgeoreg"),
            ]
        )

    if rates:
        variables.extend([f"?rate{rate}" for rate in rates])
        where_rate: List[str] = reduce(
            iconcat, [sup.operational_limit(mrid, rate) for rate in rates], []
        )
        where_list.append(sup.group_query(where_rate, command="OPTIONAL"))

    if temperatures is not None:
        variables.extend(
            [
                f"?{sup.negpos(temperature)}_{abs(temperature)}_factor"
                for temperature in temperatures
            ]
        )
        where_list.append(
            sup.group_query(
                sup.temp_correction_factors(mrid, cim, temperatures), command="OPTIONAL"
            )
        )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def connection_query(
    cim_version: int,
    rdf_types: Union[str, List[str]],
    region: Union[str, List[str]],
    sub_region: bool,
    connectivity: str,
    mrid: str,
) -> str:
    variables = [mrid, *sup.sequence_variables("t_mrid")]

    if connectivity is not None:
        variables.extend(sup.sequence_variables(connectivity))

    rdf_types = [rdf_types] if isinstance(rdf_types, str) else rdf_types
    cim_types = [sup.rdf_type_tripler(mrid, rdf_type) for rdf_type in rdf_types]

    where_list = [
        sup.combine_statements(*cim_types, group=len(cim_types) > 1, split=union_split),
        *sup.terminal_sequence_query(cim_version=cim_version, var=connectivity),
    ]

    if region is not None:
        where_list.extend(
            [
                f"{mrid} {EQUIP_CONTAINER} ?EquipmentContainer",
                "?EquipmentContainer cim:Bay.VoltageLevel ?VoltageLevel",
                f"?VoltageLevel {SUBSTATION} ?Substation",
            ]
        )
        where_list.extend(sup.region_query(region, sub_region, "Substation", "?subgeoreg"))
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))



from typing import Iterable, List, Union

import cimsparql.query_support as sup
from cimsparql.cim import EQUIP_CONTAINER, ID_OBJ, TR_WINDING


def end_number(nr: int, lock_end_number: bool) -> List[str]:
    """Lock winding mrid to specific end number

    Args:
        nr: Lock to this end number
        lock_end_number: return empty list if false
    """
    return [f"?w_mrid_{nr} cim:TransformerEnd.endNumber {nr} "] if lock_end_number else []


def terminal(mrid: str, nr: int, lock_end_number: bool = True) -> List[str]:
    """Where statements for transformer terminals"""
    return [
        *end_number(nr, lock_end_number),
        f"?w_mrid_{nr} cim:TransformerEnd.Terminal ?t_mrid_{nr}",
        f"?w_mrid_{nr} {TR_WINDING}.PowerTransformer {mrid}",
    ]


def number_of_windings(mrid: str, winding_count: int) -> str:
    """Where statement to lock query to transformers with specific number of windings"""
    variables = [mrid, "(count(distinct ?nr) as ?winding_count)"]
    where_list = [
        f"{mrid} rdf:type cim:PowerTransformer",
        f"?wwmrid {TR_WINDING}.PowerTransformer {mrid}",
        "?wwmrid cim:TransformerEnd.endNumber ?nr",
    ]
    select = sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))
    return sup.group_query(
        [select, f"group by {mrid}", f"having (?winding_count = {winding_count})"],
        command="",
        split="\n",
    )


def _market(variables: List[str], where_list: List[str], with_market: bool) -> None:
    if with_market:
        variables.extend(["?bidzone_1", "?bidzone_2"])
        where_list.extend(
            [
                "?Substation SN:Substation.MarketDeliveryPoint ?mdp",
                sup.group_query(
                    [
                        "?mdp SN:MarketDeliveryPoint.BiddingArea ?b_area",
                        "?b_area SN:BiddingArea.marketCode ?bidzone",
                        "bind(?bidzone as ?bidzone_1)",
                        "bind(?bidzone as ?bidzone_2)",
                    ],
                    command="OPTIONAL",
                ),
            ]
        )


def transformer_common(
    winding_count: int,
    mrid: str,
    name: str,
    impedance: List[str],
    variables: List[str],
    where_list: List[str],
    with_market: bool,
    region: Union[str, List[str]],
    sub_region: bool,
    rates: Iterable[str],
    network_analysis: bool,
) -> None:
    variables.extend([name, mrid, "?mrid", *sup.to_variables(impedance), "?un"])
    where_list.extend(
        [
            f"{mrid} {ID_OBJ}.name {name}",
            f"?w_mrid_1 {TR_WINDING}.ratedU ?un",
            "bind(?w_mrid_1 as ?mrid)",
            number_of_windings(mrid, winding_count),
        ]
    )
    where_list.extend(sup.predicate_list("?w_mrid_1", TR_WINDING, {z: f"?{z}" for z in impedance}))

    if with_market or region is not None:
        where_list.append(f"{mrid} {EQUIP_CONTAINER} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation", "?subgeoreg"))
        _market(variables, where_list, with_market)

    if network_analysis is not None:
        where_list.append(f"{mrid} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if rates:
        where_rate = ["?oplimitset cim:OperationalLimitSet.Terminal ?t_mrid_1"]
        for rate in rates:
            variables.append(f"?rate{rate}")
            where_rate.extend(sup.operational_limit(mrid, rate, "oplimitset"))
        where_list.append(sup.group_query(where_rate, command="OPTIONAL"))
